package multex; //Failure.java

//Change History:
//2002-07-02  Knabe  Will compile and run unchanged on JDK 1.4 and on JDK 1.1
//2002-06-27  Knabe  Will compile unchanged on JDK 1.4 and on JDK 1.1
//2002-03-26  Knabe  Example message text parameters in constructor correctly
//                   quoted
//2001-04-19  Knabe  Now its possible to redefine the checkClass-method
//2001-04-19  Knabe  Method names getCause(...) instead of cause(...),
//                   getOriginalException(...) instead of originalException(...)
//2001-04-15  Knabe  Convenience constructors with 1..9 Object-parameters added
//2001-04-14  Knabe  Again only unnamed exception parameters, not abstract
//2000-12-10  Knabe  A CauseGetter enables the MulTEx user to configure
//                   the exceptions, of which a cause can be got for reporting
//                   the causal chain.
//2000-12-08  Knabe  _cause has type Throwable rather than only Exception
//2000-10-25  Knabe  printCompactStackTrace renamed to appendCompactStackTrace
//2000-10-06  Knabe  Require default message text pattern in the exception
//2000-07-13  Knabe  Efficiently reimplemented: printCompactStackTrace
//2000-07-09  Knabe  Retrofitted to JDK 1.1: Uses Vector instead of List
//2000-06-29  Knabe  Class Exc split into Exc and Failure


/**Indirectly caused exception with parameters, causal chain and internationalizable message text.
  This class serves for the framework user as a base class for defining
  Failure-exceptions. It can be used to be thrown directly, too, but such
  exceptions cannot be handled individually nor get internationalized.
  Failure-exceptions have the following benefits: <UL>
    <LI> coherent usage always for unexpected failure of an operation
    <LI> reduce the amount of exceptions to specify in the throws-clause
         from many unexpected exceptions to only one
    <LI> chance to provide the failure of an operation with a message text,
         message parameters and a chain of the unexpected causing exceptions.
  </UL>
  The (unnamed, indexed) exception parameters are handled the same way as in
  class {@link Exc}.
  <P>
  Naming convention:
    All user-defined exception classes derived from Failure should have a name
    ending in Failure. The pattern is: <PRE>
      class OperationnameFailure extends Failure { ... }
    </PRE>
  <H2>Failure as unchecked exception?</H2>
  It can be bothering always to specify Failure in the throws-clause
  of any method, that calls other non-trivial methods. This holds especially,
  when you are converting an existing software system to usage of MulTEx.
  Thus Failure inherits the unchecked java.lang.RuntimeException.
  <P>
  So you can specify Failure, but you need not to do this.
  This is in my opinion not the cleanest way, but even without Failure
  you could not base on the assumption, that a method without a throws-clause
  does not throw any exception. Any method could throw any
  java.lang.RuntimeException, e.g. the IndexOutOfBoundsException.
  <P>
  If you want to strictly check Failure, then you can
  change the superclass of Failure to the checked java.lang.Exception instead of
  the unchecked java.lang.RuntimeException. Up to now I am not sure, what is the
  best way. Please report your experiences.
  @author Christoph Knabe, knabe@tfh-berlin.de, TFH Berlin, 1999-2000 Copyrighted
*/

public class Failure extends RuntimeException
implements MultexException {


public static final String className = "Failure";

/**The string used to delimit the stack trace part of a causing exception from
  the following stack trace of the caused Failure-exception.
*/
public static final String wasCausing = "WAS CAUSING:";

/**Convenience constructor null
  @see #Failure(String, Throwable, Object[])
*/
public Failure(
  final String i_defaultMessageTextPattern, final Throwable i_cause
){
  this(i_defaultMessageTextPattern, i_cause, (Object[])null);
}

/**Convenience constructor 0
  @see #Failure(String, Throwable, Object[])
*/
public Failure(
  final String i_defaultMessageTextPattern, final Throwable i_cause,
  final Object i_object0
){
  this(i_defaultMessageTextPattern, i_cause, new Object[]{
    i_object0
  });
}

/**Convenience constructor 1
  @see #Failure(String, Throwable, Object[])
*/
public Failure(
  final String i_defaultMessageTextPattern, final Throwable i_cause,
  final Object i_object0, final Object i_object1
){
  this(i_defaultMessageTextPattern, i_cause, new Object[]{
    i_object0, i_object1
  });
}

/**Convenience constructor 2
  @see #Failure(String, Throwable, Object[])
*/
public Failure(
  final String i_defaultMessageTextPattern, final Throwable i_cause,
  final Object i_object0, final Object i_object1, final Object i_object2
){
  this(i_defaultMessageTextPattern, i_cause, new Object[]{
    i_object0, i_object1, i_object2
  });
}

/**Convenience constructor 3
  @see #Failure(String, Throwable, Object[])
*/
public Failure(
  final String i_defaultMessageTextPattern, final Throwable i_cause,
  final Object i_object0, final Object i_object1, final Object i_object2,
  final Object i_object3
){
  this(i_defaultMessageTextPattern, i_cause, new Object[]{
    i_object0, i_object1, i_object2, i_object3
  });
}

/**Convenience constructor 4
  @see #Failure(String, Throwable, Object[])
*/
public Failure(
  final String i_defaultMessageTextPattern, final Throwable i_cause,
  final Object i_object0, final Object i_object1, final Object i_object2,
  final Object i_object3, final Object i_object4
){
  this(i_defaultMessageTextPattern, i_cause, new Object[]{
    i_object0, i_object1, i_object2, i_object3, i_object4
  });
}

/**Convenience constructor 5
  @see #Failure(String, Throwable, Object[])
*/
public Failure(
  final String i_defaultMessageTextPattern, final Throwable i_cause,
  final Object i_object0, final Object i_object1, final Object i_object2,
  final Object i_object3, final Object i_object4, final Object i_object5
){
  this(i_defaultMessageTextPattern, i_cause, new Object[]{
    i_object0, i_object1, i_object2, i_object3, i_object4, i_object5
  });
}

/**Convenience constructor 6
  @see #Failure(String, Throwable, Object[])
*/
public Failure(
  final String i_defaultMessageTextPattern, final Throwable i_cause,
  final Object i_object0, final Object i_object1, final Object i_object2,
  final Object i_object3, final Object i_object4, final Object i_object5,
  final Object i_object6
){
  this(i_defaultMessageTextPattern, i_cause, new Object[]{
    i_object0, i_object1, i_object2, i_object3, i_object4, i_object5, i_object6
  });
}

/**Convenience constructor 7
  @see #Failure(String, Throwable, Object[])
*/
public Failure(
  final String i_defaultMessageTextPattern, final Throwable i_cause,
  final Object i_object0, final Object i_object1, final Object i_object2,
  final Object i_object3, final Object i_object4, final Object i_object5,
  final Object i_object6, final Object i_object7
){
  this(i_defaultMessageTextPattern, i_cause, new Object[]{
    i_object0, i_object1, i_object2, i_object3, i_object4, i_object5, i_object6,
    i_object7
  });
}

/**Convenience constructor 8
  @see #Failure(String, Throwable, Object[])
*/
public Failure(
  final String i_defaultMessageTextPattern, final Throwable i_cause,
  final Object i_object0, final Object i_object1, final Object i_object2,
  final Object i_object3, final Object i_object4, final Object i_object5,
  final Object i_object6, final Object i_object7, final Object i_object8
){
  this(i_defaultMessageTextPattern, i_cause, new Object[]{
    i_object0, i_object1, i_object2, i_object3, i_object4, i_object5, i_object6,
    i_object7, i_object8
  });
}

/**Convenience constructor 9
  @see #Failure(String, Throwable, Object[])
*/
public Failure(
  final String i_defaultMessageTextPattern, final Throwable i_cause,
  final Object i_object0, final Object i_object1, final Object i_object2,
  final Object i_object3, final Object i_object4, final Object i_object5,
  final Object i_object6, final Object i_object7, final Object i_object8,
  final Object i_object9
){
  this(i_defaultMessageTextPattern, i_cause, new Object[]{
    i_object0, i_object1, i_object2, i_object3, i_object4, i_object5, i_object6,
    i_object7, i_object8, i_object9
  });
}

/**Constructs a Failure-exception, giving complete diagnostic information.
  The diagnostic information consists of <OL>
    <LI>the default message text pattern
    <LI>the causing Throwable object for providing the diagnostics causer chain
    <LI>exception parameters as an polymorphic Object[], which can be inserted
        into the message text pattern by placeholders {0} ... {9}.
  </OL>
* <P>Example of defining an exception with text, cause and parameters: <PRE>
*   public static class CopyFailure extends Failure {
*     public CopyFailure(
*       final Throwable i_cause,
*       final String    i_inFileName,
*       final String    i_outFileName
*     ){
*       super("File ''{0}'' could not be copied to ''{1}''",
*         i_cause, new Object[]{i_inFileName, i_outFileName}
*       );
*     }
*   }//CopyFailure
* </PRE>
  Instead of creating an Object[] yourself you can use the corresponding
  convenience constructors.
*/
public Failure(
  final String i_defaultMessageTextPattern,
  final Throwable i_cause, final Object[] i_parameters
){
  _defaultMessageTextPattern = i_defaultMessageTextPattern;
  _cause = i_cause;
  _parameters = i_parameters;
  checkClass();
}

public String getDefaultMessageTextPattern(){
  return _defaultMessageTextPattern;
}

public Object[] getParameters(){return _parameters;}

/**Result: The Throwable object, which directly caused the program to throw this
  Failure-exception. I.e. the Throwable object, which was redefined into
  this Failure object,
* if there is one, otherwise null.
*/
public final Throwable getCause(){return _cause;}

/** Result: the directly causing Throwable object of i_throwable,
* if there is one, otherwise null.
*/
public final static Throwable getCause(final Throwable i_throwable){
  return i_throwable instanceof Failure ? ((Failure)i_throwable).getCause()
    : _causeGetter.getCause(i_throwable)
  ;
}

/**Result: The deepest Throwable object, which indirectly caused the
  program to throw i_throwable, if there is one, otherwise i_throwable
  itself.
  @see #getCause(Throwable)
*/
public final static Throwable getOriginalException(final Throwable i_throwable){
  for(Throwable result = i_throwable;;){
    final Throwable cause = getCause(result);
    if(cause==null){return result;}
  }
}

/**Sets a CauseGetter to be used by method {@link #getCause(Throwable)}
  in order to get the cause of a Throwable object, which is not a subclass
  of Failure. Default is {@link Jdk1_4CauseGetter}
*/
public final static void setCauseGetter(final CauseGetter i_causeGetter){
  _causeGetter = i_causeGetter;
}

public String getMessage(){return Exc.getUserInformation(this);}

/** Prints the chained, raw stack traces of this <code>Failure</code>
  object to the standard error stream
  <A HREF="http://java.sun.com/j2se/1.3/docs/api/java/lang/System.html#err">System.err</A>.
  @see #printStackTrace(java.io.PrintWriter)
*/
public void printStackTrace(){
  printStackTrace(System.err);
}

/** Prints the chained, raw stack traces of this <code>Failure</code>
  object to io_printer.
* @see #printStackTrace(java.io.PrintWriter)
*/
public void printStackTrace(final java.io.PrintStream io_printer) {
  printStackTrace(new java.io.PrintWriter(io_printer, true));
}

/** Prints the chained, raw stack traces of this <code>Failure</code>
  object to io_printer.
  Depending on the version of the Java Runtime Environment we get for<UL>
    <LI>1.4 or higher: the standard Throwable.printStackTrace() with its
      own built-in exception chain. Users can call this method directly.
    <LI>1.1 up to 1.3: The following MulTEx stack trace:
  The output contains the name, message parameters and stack trace first of
  the causing <code>Throwable</code> object, then of this <code>Failure</code>
  object to the specified print writer. Users should usually not call this
  method directly, but {@link Msg#printStackTrace} instead, in order to avoid
  redundant and useless lines in the stack traces.
*/
public void printStackTrace(final java.io.PrintWriter io_printer) {
  if(_jreHasExceptionChaining){ //From JRE 1.4 there the standard
    //Throwable.printStackTrace will do the exception chain printing. Thus it is
    //unnecessary to recurse into the causal chain here
  }else{ //JRE 1.1 - 1.3: Follow chain ourselves, report lowest cause first:
    //simple efficient code: leaves all redundant lines:
    if(_cause!=null){
      _cause.printStackTrace(io_printer);
      io_printer.println(wasCausing);
    }
  }
  super.printStackTrace(io_printer);
  io_printer.flush();
}


/**Appends the stack trace of i_throwable and all its chained cause exceptions
  to io_buffer. All redundant location lines therein are suppressed.
  Usually you should not directly call this method, but a printStackTrace-method in
  class Msg.
  @param io_buffer Where to append the compactified stack trace
  @param i_reportee The exception to report, possibly containing a chain of
    nested cause exceptions
  @param i_here An dummy exception, which should have been newly created there,
    where the reporting for i_reportee is invoked; No longer interpreted.
  @see Msg#printStackTrace(StringBuffer,Throwable)
  @deprecated Obsolete from multex 2, deprecated from 3g.
*/
public static void appendCompactStackTrace(
  final StringBuffer io_buffer, final Throwable i_reportee ,final Exception i_here
){
  appendCompactStackTrace(io_buffer, i_reportee, lineSeparator);	
}

/**Appends the stack trace of i_throwable and all its chained cause exceptions
  to io_buffer. All redundant location lines therein are suppressed.
  Usually you should not directly call this method, but a printStackTrace-method in
  class Msg.
  @see Msg#printStackTrace(StringBuffer,Throwable,String)
*/
/*package*/ static void appendCompactStackTrace(
  final StringBuffer io_destination, final Throwable i_throwable, 
  final String i_lineSeparator
){
  //prepare out writer:
  final java.io.StringWriter sw = new java.io.StringWriter();
  final java.io.PrintWriter out = new java.io.PrintWriter(sw, true);
  //final java.io.PrintWriter err = new java.io.PrintWriter(System.err, true);

  //collect stack traces onto out writer:
  i_throwable.printStackTrace(out);

  /*2003-05-07 Knabe: obsolete
  final boolean shortenLastTrace = false;
  if(shortenLastTrace){
    out.println(wasCausing);
    i_here.printStackTrace(out);
  }
  */
  final String tracesString = sw.toString();
  //System.err.println("tracesString:");System.err.println(tracesString);//debug
  final java.util./*String*/Vector tracesLines
  = _splitIntoLines(tracesString);

  if(!_jreHasExceptionChaining){ 
  	//standard printStackTrace() left redundant lines in stack traces
  	//of causing exceptions!
  	nullifyRedundantTraceTrailLines(tracesLines);
  }
  _printLines(io_destination, tracesLines, i_lineSeparator);
}//appendCompactStackTrace

/**Nullifies the redundant lines ending each of the contained stack traces.
 * Since the stack trace of an exception and its cause have much in common,
 * there is a big redundancy. From JDK 1.4 this redundancy is eliminated by
 * the standard printStackTrace(). Before JDK 1.4 the redundancy elimination
 * is the task of this method.
 * Looks up the end locations of each of the stack traces, which are separated 
 * by the word contained in wasCausing. The goin downwards from each of these
 * end locations nullifies the redundant lines.
 * */
private static void nullifyRedundantTraceTrailLines(
	final java.util.Vector io_tracesLines
) {
	  final int[] traceEndIndices = _getTraceEndIndices(io_tracesLines);
	  //_println("traceEndIndices", traceEndIndices); //debug
	  for(int i=1; i<traceEndIndices.length; i++){
	    _nullifyCommonTraceTail(
	      io_tracesLines, traceEndIndices[i-1], traceEndIndices[i]
	    );
	  }
	  /*2003-05-07 Knabe: obsolete
	  if(shortenLastTrace){
	    _shortenLastTrace(
	      tracesLines, traceEndIndices[traceEndIndices.length-2]+1
	    );
	  }
	  */
}//nullifyRedundantTraceTrailLines

/**Splits i_string into internized line strings by BufferedReader.readln().
  @return The lines in the sequence of storage in i_string as a Vector of
          String-s.
*/
private static java.util./*String*/Vector _splitIntoLines(final String i_string){
  final java.io.BufferedReader reader = new java.io.BufferedReader(
    new java.io.StringReader(i_string)
  );
  final java.util./*String*/Vector result = new java.util.Vector();
  for(;;){
    String line; //Warum geht es hier nicht final?
    try{line = reader.readLine();}
      catch(java.io.IOException ex){line=null;}
    //end try
    if(line==null){break;}
    result.addElement(line.intern());
  }
  return result;
}

/**Result: The indices of the elements of i_tracesLines, which each represent
  the last line of a trace. Each line immediately before an occurence of the
  separator String wasCausing in i_tracesLines is considered as a trace end
  line. Additionally the last line of i_tracesLines is considered as a trace
  end line.
  Require: In i_traces are several stack traces, split into lines; the traces
    are separated by the String wasCausing.
*/
private static int[] _getTraceEndIndices(final java.util.Vector i_tracesLines){
  final int numberOfTraces
  = _numberOfOccurences(i_tracesLines, wasCausing) + 1;
  final int[] result = new int[numberOfTraces];
  final int size = i_tracesLines.size();
  int resultIndex = 0;
  for(int i=0; i<size; i++){
    if(i_tracesLines.elementAt(i) == wasCausing){
      result[resultIndex++] = i-1; //line before wasCausing
    }
  }
  result[resultIndex++] = size-1; //last line
  return result;
}

/**Result: The number, how often i_object occurs in i_vector,
  measured by the identity comparison
*/
private static int _numberOfOccurences
(final java.util.Vector i_vector, Object i_object){
  final int size = i_vector.size();
  int result = 0;
  for(int i=0; i<size; i++){
    if(i_vector.elementAt(i) == i_object){result++;}
  }
  return result;
}

/**Nullifies redundant subsequent entries in io_vector.
  Nullifies all entries in io_vector
  starting from i_nullifyEndIndex inclusively downwards,
  which are identically contained in the same
  sequence as entries starting from i_compareEndIndex inclusively downwards.
  The lowest such redundant entry will not be nullified.
*/
private static void _nullifyCommonTraceTail(
  final java.util.Vector io_vector,
  final int i_nullifyEndIndex, final int i_compareEndIndex
){
  for(int ni=i_nullifyEndIndex, ci=i_compareEndIndex; ni>=0;ni--,ci--){
    if(io_vector.elementAt(ni)!=io_vector.elementAt(ci)){break;}
    io_vector.setElementAt(null, ni); //leaves no redundant lines
    /*de-experimented 01-04-13. Leaves one redundant line in lower stack trace:
    if(ni<i_nullifyEndIndex){
      io_vector.setElementAt(null, ni+1);
    }
    */
  }
}//_nullifyCommonTraceTail

/*Commented out 2003-05-07: obsolete
 *  
/**Shortens the Vector io_tracesLines by setting the element at index
  i_fromIndex to the abbreviation symbol "at ..." and nullifying all
  elements after this.
* /
private static void _shortenLastTrace(
  final java.util.Vector io_tracesLines, final int i_fromIndex
){
  io_tracesLines.setElementAt("\tat ...", i_fromIndex);
  final int size = io_tracesLines.size();
  for(int i=i_fromIndex+1; i<size; i++){
    io_tracesLines.setElementAt(null, i);
  }
}
*/

/**Appends the non-null elements of i_lines each as a line to io_buffer,
 * each followed by i_lineSeparator
 **/
private static void _printLines(
  final StringBuffer    io_buffer,
  final java.util.Vector i_lines,
  final String           i_lineSeparator
){
  final int size = i_lines.size();
  for(int i=0; i<size; i++){
    final Object line = i_lines.elementAt(i);
    if(line!=null){io_buffer.append(line);io_buffer.append(i_lineSeparator);}
  }
}

/**Returns the line separator for the actual platform.
  This is the line separator for the platform the Java virtual machine is
  running on. It is the value of the system property 'line.separator'.
*/
public static final String lineSeparator
= System.getProperty("line.separator");

/**Checks that the class of this object is OK.
  In the moment this means, that its name ends in {@link #className}
*/
protected void checkClass(){
  Exc.checkClass(this, className);
}

private /*Java1.2:final*/ String _defaultMessageTextPattern;

/**Unnamed parameters of this exception.*/
private /*Java1.2:final*/ Object[] _parameters;

/**The Throwable object, that caused this Failure-exception
  @see #getCause()
*/
private /*Java1.2:final*/ Throwable _cause;

/**The version number of the Java Runtime Environment, on which we are running*/
private static final String  _jreVersion = System.getProperty("java.version");

/**Is true, if the JRE-version number is >= 1.4*/
private static final boolean _jreHasExceptionChaining
= _jreVersion.compareTo("1.4") >= 0;

/**Returns true, if class java.lang.Throwable has an operation getCause() in this Java Runtime Environment*/
public static boolean jreHasExceptionChaining(){return _jreHasExceptionChaining;}

/**Getter for the cause of an exception, by default the ReflectionCauseGetter*/
private static CauseGetter _causeGetter = new ReflectionCauseGetter();
/* replaced by using the ReflectionCauseGetter:
{
    try{
        _causeGetter= _jreHasExceptionChaining
        ? (CauseGetter)(Class.forName("multex.Jdk1_4CauseGetter").newInstance())
        //here we had to use reflection instead of simply saying
        //  (CauseGetter)new Jdk1_4CauseGetter()
        //in order to be able to compile this class on JDK 1.1, too

        : (CauseGetter)new Jdk1_1CauseGetter()
        ;
    }catch(Exception ex){
        System.err.println("MulTEx-CLASS " + className + ": UNEXPECTED EXCEPTION INITIALIZING ME:");
        throw new ExceptionInInitializerError(ex);
    }
}
*/


}//Failure