package multex; //Jdk1_4CauseGetter.java

//History:
//2002-06-24  Knabe  Re-added java.sql.SQLException, added java.sql.SQLWarning
//2002-04-29  Knabe  Simpler Approach with built-in causes in JDK 1.4
//2001-04-19  Knabe  Moved here from inner class Failure.Jdk1_1CauseGetter

/**Describes, how to get the causing exception of any exception in JDK 1.4
  with an accessible causing exception.
  For the exceptions in java.sql calls their method getNextException() or
  getNextWarning(), respectively. For all other exceptions uses the standard
  cause getter Throwable.getcause().
  @deprecated Compiles only on JDK 1.4 and is then the most efficient solution
    for getting the cause of an exception. As of forward/backward compatibility
    issues it is now preferable to use the {@link ReflectionCauseGetter}.
    In the future, when only standard exception chaining  la Throwable.getCause()
    will be used, then this CauseGetter will be the most efficient.
*/
public class Jdk1_4CauseGetter implements CauseGetter {
  public Throwable getCause(final Throwable i_throwable){
    if(i_throwable instanceof java.sql.SQLWarning){ //MulTEx 3b: 2002-06-24 Knabe
      return ((java.sql.SQLWarning)i_throwable).getNextWarning();
    }else if(i_throwable instanceof java.sql.SQLException){ //MulTEx 1c: 2001-02-27 Knabe
      return ((java.sql.SQLException)i_throwable).getNextException();
    }
    return i_throwable.getCause(); //compiles only from JDK 1.4
    //Do not compile this class for usage on JDKs up to 1.3
  }
}
